"""IMF FSI Data Set Helpers."""

from typing import Optional


def load_fsi_symbols() -> dict:
    """Load IMF FSI symbols."""
    # pylint: disable=import-outside-toplevel
    from openbb_imf.utils.constants import load_symbols

    return load_symbols("FSI")


def validate_symbols(symbols) -> str:
    """Validate the IMF FSI symbols."""
    # pylint: disable=import-outside-toplevel
    from warnings import warn  # noqa
    from openbb_core.app.model.abstract.error import OpenBBError
    from openbb_imf.utils.constants import FSI_PRESETS

    fsi_symbols = load_fsi_symbols()

    if isinstance(symbols, str):
        symbols = symbols.split(",")
    elif not isinstance(symbols, list):
        raise OpenBBError("Invalid symbols list.")

    new_symbols: list = []

    for symbol in symbols:
        if symbol in FSI_PRESETS:
            return symbol
        if symbol.upper() not in fsi_symbols:
            warn(f"Unsupported IMF FSI symbol: {symbol}")
        new_symbols.append(symbol.upper())

    return "+".join(new_symbols) if len(new_symbols) > 1 else new_symbols[0]


# pylint: disable=too-many-branches,too-many-statements,too-many-locals
async def _get_fsi_data(**kwargs) -> list[dict]:  # noqa:PLR0912
    """Get IMF FSI data.
    This function is not intended to be called directly,
    but through the `ImfEconomicIndicatorsFetcher` class.
    """
    # pylint: disable=import-outside-toplevel
    from datetime import date  # noqa
    from openbb_core.app.model.abstract.error import OpenBBError
    from openbb_imf.utils.helpers import get_data
    from openbb_imf.utils.irfcl_helpers import load_country_to_code_map
    from openbb_imf.utils import constants
    from pandas import to_datetime
    from pandas.tseries import offsets

    countries = kwargs.get("country", "")
    countries = (
        "" if countries == "all" else countries.replace(",", "+") if countries else ""
    )
    frequency = constants.FREQUENCY_DICT.get(kwargs.get("frequency", "quarter"), "Q")
    start_date = kwargs.get("start_date", "")
    end_date = kwargs.get("end_date", "")
    all_symbols = load_fsi_symbols()
    core_only = [k for k, v in all_symbols.items() if v.get("table") == "fsi_core"]
    encouraged_set = [
        k for k, v in all_symbols.items() if v.get("table") == "fsi_encouraged_set"
    ]
    indicator = kwargs.get("symbol")
    indicators = (
        "+".join(core_only if indicator == "fsi_core" else encouraged_set)
        if indicator in ["fsi_core", "fsi_encouraged_set"]
        else (
            ""
            if indicator
            in ["fsi_other", "fsi_all", "fsi_core_underlying", "fsi_balance_sheets"]
            else validate_symbols(indicator)
        )
    )

    if not indicators and not countries:
        raise OpenBBError(
            "All countries not supported for this group of indicators. Please supply a country."
        )

    if not start_date and not end_date and not countries:
        start_date = (
            date.today().replace(year=date.today().year - 1).strftime("%Y-%m-%d")
        )

    if start_date and not end_date:
        end_date = date.today().strftime("%Y-%m-%d")

    # Adjust the dates to the date relative to frequency.
    # The API does not accept arbitrary dates, so we need to adjust them.
    if start_date:
        start_date = to_datetime(start_date)
        if frequency == "Q":
            start_date = offsets.QuarterBegin(startingMonth=1).rollback(start_date)
        elif frequency == "A":
            start_date = offsets.YearBegin().rollback(start_date)
        else:
            start_date = offsets.MonthBegin().rollback(start_date)
        start_date = start_date.strftime("%Y-%m-%d")

    if end_date:
        end_date = to_datetime(end_date)
        if frequency == "Q":
            end_date = offsets.QuarterEnd().rollforward(end_date)
        elif frequency == "A":
            end_date = offsets.YearEnd().rollforward(end_date)
        else:
            end_date = offsets.MonthEnd().rollforward(end_date)
        end_date = end_date.strftime("%Y-%m-%d")

    date_range = (
        f"?startPeriod={start_date}&endPeriod={end_date}"
        if start_date and end_date
        else ""
    )
    base_url = "http://dataservices.imf.org/REST/SDMX_JSON.svc/"
    key = f"CompactData/FSI/{frequency}.{countries}.{indicators}"
    url = f"{base_url}{key}{date_range}"
    series = await get_data(url)
    data: list = []
    all_symbols = load_fsi_symbols()

    if indicator in ["fsi_core", "fsi_encouraged_set"]:
        all_symbols = {
            k: v for k, v in all_symbols.items() if v.get("table") == indicator
        }
    elif indicator == "fsi_core_underlying":
        all_symbols = {
            k: v
            for k, v in all_symbols.items()
            if "core set" in v.get("title", "").lower() and v.get("unit") != "Percent"
        }
    elif indicator == "fsi_balance_sheets":
        all_symbols = {
            k: v
            for k, v in all_symbols.items()
            if "balance sheets" in v.get("title", "").lower()
        }
    elif indicator == "fsi_other":
        all_symbols = {
            k: v
            for k, v in all_symbols.items()
            if "Additional FSIs" in v.get("title", "")
        }

    country_map_dict = {
        v: k.replace("_", " ").title().replace("Ecb", "ECB")
        for k, v in load_country_to_code_map().items()
    }
    # Iterate over the series to extract observations and map the metadata.
    for s in series:
        if "Obs" not in s:
            continue
        meta = {
            k.replace("@", "").lower(): (
                constants.UNIT_MULTIPLIERS_MAP.get(str(v), v)
                if k == "@UNIT_MULT"
                else v
            )
            for k, v in s.items()
            if k != "Obs"
        }
        _symbol = meta.get("indicator")
        _parent: Optional[str] = None
        _order: Optional[str] = None
        _level: Optional[str] = None
        _table: Optional[str] = None
        _title: Optional[str] = None
        _unit: Optional[str] = None

        if _symbol not in all_symbols:
            continue

        _table = all_symbols.get(_symbol, {}).get("table", "")
        if _table and indicator == "fsi_core_underlying":
            _table = _table.replace("fsi_other", "fsi_core_underlying")  # type: ignore
        _parent = all_symbols.get(_symbol, {}).get("parent", "")
        _order = all_symbols.get(_symbol, {}).get("order", "")
        _level = all_symbols.get(_symbol, {}).get("level", "")
        _title = all_symbols.get(_symbol, {}).get("title", "")
        if _title:
            _title = " - ".join(_title.split(", ")[1:-1]).replace(
                "Financial Soundness Indicators - ", ""
            )
        _unit = all_symbols.get(_symbol, {}).get("unit", "")

        _data = s.pop("Obs", [])

        if isinstance(_data, dict):
            _data = [_data]

        for d in _data:
            _date = d.pop("@TIME_PERIOD", None)
            val = d.pop("@OBS_VALUE", None)
            _ = d.pop("@OBS_STATUS", None)
            if not val:
                continue
            val = float(val)
            if _unit and _unit.lower() == "percent":
                val = val / 100
            if _date:
                offset = (
                    offsets.QuarterEnd
                    if "Q" in _date
                    else offsets.YearEnd if len(str(_date)) == 4 else offsets.MonthEnd
                )
                _date = to_datetime(_date)
                _date = _date + offset(0)
                _date = _date.strftime("%Y-%m-%d")
            vals = {
                k: v
                for k, v in {
                    "date": _date,
                    "table": _table,
                    "symbol": _symbol,
                    "parent": _parent if _parent else _symbol,
                    "level": _level,
                    "order": _order,
                    "country": country_map_dict.get(
                        meta.get("ref_area"), meta.get("ref_area")
                    ),
                    "reference_sector": None,
                    "title": _title,
                    "value": val,
                    "unit": (
                        _unit.upper() if _unit and _unit in ["usd", "eur"] else _unit
                    ),
                    "scale": (
                        "Basis Points"
                        if _unit and _unit.lower() == "percent"
                        else meta.get("unit_mult")
                    ),
                }.items()
                if v
            }

            if vals.get("value") and vals.get("date"):
                d.update(vals)
                data.append(d)

    if not data:
        raise OpenBBError(f"No data found for, '{indicator}', in, '{countries}'.")

    return data



"""IMF Available Indicators."""

# pylint: disable=unused-argument

from typing import Any, Optional, Union

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.available_indicators import (
    AvailableIndicatorsData,
    AvailableIndicesQueryParams,
)
from pydantic import Field


class ImfAvailableIndicatorsQueryParams(AvailableIndicesQueryParams):
    """IMF Available Indicators Query Parameters."""

    __json_schema_extra__ = {"query": {"multiple_items_allowed": True}}

    query: Optional[str] = Field(
        default=None,
        description="The query string to search through the available indicators."
        + " Use semicolons to separate multiple terms.",
    )


class ImfAvailableIndicatorsData(AvailableIndicatorsData):
    """IMF Available Indicators Data."""

    __alias_dict__ = {
        "symbol_root": "parent",
        "description": "title",
    }
    dataset: Optional[str] = Field(
        default=None,
        description="The IMF dataset associated with the symbol.",
    )
    table: Optional[str] = Field(
        default=None,
        description="The name of the table associated with the symbol.",
    )
    level: Optional[int] = Field(
        default=None,
        description="The indentation level of the data, relative to the table and symbol_root",
    )
    order: Optional[Union[int, float]] = Field(
        default=None,
        description="Order of the data, relative to the table.",
    )
    children: Optional[str] = Field(
        default=None,
        description="The symbol of the child data, if any.",
    )
    unit: Optional[str] = Field(
        default=None,
        description="The unit of the data.",
    )


class ImfAvailableIndicatorsFetcher(
    Fetcher[ImfAvailableIndicatorsQueryParams, list[ImfAvailableIndicatorsData]]
):
    """IMF Available Indicators Fetcher."""

    @staticmethod
    def transform_query(params: dict[str, Any]) -> ImfAvailableIndicatorsQueryParams:
        """Transform the query."""
        return ImfAvailableIndicatorsQueryParams(**params)

    @staticmethod
    def extract_data(
        query: ImfAvailableIndicatorsQueryParams,
        credentials: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> list[dict]:
        """Fetch the data."""
        # pylint: disable=import-outside-toplevel
        from numpy import nan
        from openbb_core.provider.utils.errors import EmptyDataError
        from openbb_imf.utils.constants import load_symbols
        from pandas import DataFrame, Series

        try:
            all_symbols = load_symbols("all")
        except OpenBBError as e:
            raise OpenBBError(f"Failed to load IMF symbols static file: {e}") from e

        terms = [term.strip() for term in query.query.split(";")] if query.query else []

        df = (
            DataFrame(all_symbols)
            .T.reset_index()
            .rename(columns={"index": "symbol"})
            .replace({nan: None})
        )

        if not terms:
            records = df.to_dict(orient="records")
        else:
            combined_mask = Series([True] * len(df))
            for term in terms:
                mask = df.apply(
                    lambda row, term=term: row.astype(str).str.contains(
                        term, case=False, regex=True, na=False
                    )
                ).any(axis=1)
                combined_mask &= mask

            matches = df[combined_mask]

            if matches.empty:
                raise EmptyDataError("No results found for the provided query.")

            records = matches.to_dict(orient="records")

        return records

    @staticmethod
    def transform_data(
        query: ImfAvailableIndicatorsQueryParams,
        data: list[dict],
        **kwargs: Any,
    ) -> list[ImfAvailableIndicatorsData]:
        """Transform the data."""
        return [ImfAvailableIndicatorsData.model_validate(d) for d in data]



"""IMF Helper Utilities."""


async def get_data(url: str) -> list[dict]:
    """Get data from the IMF API."""
    # pylint: disable=import-outside-toplevel

    from aiohttp.client_exceptions import ContentTypeError  # noqa
    from json.decoder import JSONDecodeError
    from openbb_core.provider.utils.helpers import amake_request
    from openbb_core.app.model.abstract.error import OpenBBError

    try:
        response = await amake_request(url, timeout=20)
    except (JSONDecodeError, ContentTypeError) as e:
        raise OpenBBError(
            "Error fetching data; This might be rate-limiting. Try again later."
        ) from e

    if "ErrorDetails" in response:
        raise OpenBBError(
            f"{response['ErrorDetails'].get('Code')} -> {response['ErrorDetails'].get('Message')}"  # type: ignore
        )

    series = response.get("CompactData", {}).get("DataSet", {}).pop("Series", {})  # type: ignore

    if not series:
        raise OpenBBError(f"No time series data found -> {response}")

    # If there is only one series, they ruturn a dict instead of a list.
    if series and isinstance(series, dict):
        series = [series]

    return series



"""IMF Constants."""

# pylint: disable=line-too-long
# flake8: noqa: E501

FSI_PRESETS = [
    "fsi_core",
    "fsi_core_underlying",
    "fsi_other",
    "fsi_encouraged_set",
    "fsi_balance_sheets",
    "fsi_all",
]

IRFCL_HEADLINE = "RAF_USD,RAFA_USD,RAFAFX_USD,RAOFA_USD,RAPFA_USD,RAFAIMF_USD,RAFASDR_USD,RAFAGOLD_USD,RACFA_USD,RAMDCD_USD,RAMFIFC_USD,RAMSR_USD"

RESERVE_ASSETS_AND_OTHER_FX_ASSETS = "RAF_USD,RAFA_USD,RAFAFX_USD,RAFAFXS_USD,RAFAFXSI_USD,RAFAFXCD_USD,RAFAFXCDN_USD,RAFAFXCDBI_USD,RAFAFXCDBIA_USD,RAFAFXCDBO_USD,RAFAFXCDBOA_USD,RAFAIMF_USD,RAFASDR_USD,RAFAGOLD_USD,RAFAGOLDV_OZT,RAFAO_USD,RAFAOF_USD,RAFAOL_USD,RAFAOO_USD,RAOFA_USD,RAOFAS_USD,RAOFAD_USD,RAOFAL_USD,RAOFAF_USD,RAOFAG_USD"

PREDETERMINED_DRAINS_ON_FX_ASSETS = "RAPFA_USD,RAPFALSD_USD,RAPFALSD_1M_USD,RAPFALSD_1M_3M_USD,RAPFALSD_3M_1Y_USD,RAPFALSDOP_USD,RAPFALSDOP_1M_USD,RAPFALSDOP_1M_3M_USD,RAPFALSDOP_3M_1Y_USD,RAPFALSDOI_USD,RAPFALSDOI_1M_USD,RAPFALSDOI_1M_3M_USD,RAPFALSDOI_3M_1Y_USD,RAPFALSDIP_USD,RAPFALSDIP_1M_USD,RAPFALSDIP_1M_3M_USD,RAPFALSDIP_3M_1Y_USD,RAPFALSDII_USD,RAPFALSDII_1M_USD,RAPFALSDII_1M_3M_USD,RAPFALSDII_3M_1Y_USD,RAPFAFFS_USD,RAPFAFFS_1M_USD,RAPFAFFS_1M_3M_USD,RAPFAFFS_3M_1Y_USD,RAPFAFFL_USD,RAPFAFFL_1M_USD,RAPFAFFL_1M_3M_USD,RAPFAFFL_3M_1Y_USD,RAPFAO_USD,RAPFAO_1M_USD,RAPFAO_1M_3M_USD,RAPFAO_3M_1Y_USD,RAPFAOOR_USD,RAPFAOOR_1M_USD,RAPFAOOR_1M_3M_USD,RAPFAOOR_3M_1Y_USD,RAPFAOIRR_USD,RAPFAOIRR_1M_USD,RAPFAOIRR_1M_3M_USD,RAPFAOIRR_3M_1Y_USD,RAPFAOOC_USD,RAPFAOOC_1M_USD,RAPFAOOC_1M_3M_USD,RAPFAOOC_3M_1Y_USD,RAPFAOIC_USD,RAPFAOIC_1M_USD,RAPFAOIC_1M_3M_USD,RAPFAOIC_3M_1Y_USD,RAPFAOOP_USD,RAPFAOOP_1M_USD,RAPFAOOP_1M_3M_USD,RAPFAOOP_3M_1Y_USD,RAPFAOIR_USD,RAPFAOIR_1M_USD,RAPFAOIR_1M_3M_USD,RAPFAOIR_3M_1Y_USD,RAFA_RAPFA_RO"

CONTINGENT_DRAINS_FX_ASSETS = "RACFA_USD,RACFAL_USD,RACFAL_1M_USD,RACFAL_1M_3M_USD,RACFAL_3M_1Y_USD,RACFALG_USD,RACFALG_1M_USD,RACFALG_1M_3M_USD,RACFALO_USD,RACFALO_1M_USD,RACFALO_1M_3M_USD,RACFALO_3M_1Y_USD,RACFAS_USD,RACFACB_USD,RACFACB_1M_USD,RACFACB_1M_3M_USD,RACFACB_3M_1Y_USD,RACFACBA_USD,RACFACBA_1M_USD,RACFACBA_1M_3M_USD,RACFACBA_3M_1Y_USD,RACFACBAOI_USD,RACFACBAOI_1M_USD,RACFACBAOI_1M_3M_USD,RACFACBAON_USD,RACFACBAON_1M_USD,RACFACBAON_1M_3M_USD,RACFACBAON_3M_1Y_USD,RACFACBABIS_USD,RACFACBAIMF_1M_USD,RACFACBAIMF_1M_3M_USD,RACFACBAIMF_3M_1Y_USD,RACFACBAIMF_USD,RACFACBFIR_USD,RACFACBFIR_1M_USD,RACFACBFIR_1M_3M_USD,RACFACBFIR_3M_1Y_USD,RACFACBFIO_USD,RACFACBFIO_1M_USD,RACFACBFIO_1M_3M_USD,RACFACBFIO_3M_1Y_USD,RACFACT_USD,RACFACT_1M_USD,RACFACT_1M_3M_USD,RACFACT_3M_1Y_USD,RACFACTA_USD,RACFACTA_1M_USD,RACFACTA_1M_3M_USD,RACFACTA_3M_1Y_USD,RACFACTAOI_USD,RACFACTAOI_1M_USD,RACFACTAOI_1M_3M_USD,RACFACTAOI_3M_1Y_USD,RACFACTAON_USD,RACFACTAON_1M_USD,RACFACTAON_1M_3M_USD,RACFACTAON_3M_1Y_USD,RACFACTABIS_USD,RACFACTABIS_1M_USD,RACFACTABIS_1M_3M_USD,RACFACTABIS_3M_1Y_USD,RACFACTAIMF_USD,RACFACTAIMF_1M_USD,RACFACTFIR_USD,RACFACTFIR_1M_USD,RACFACTFIR_1M_3M_USD,RACFACTFIR_3M_1Y_USD,RACFACTFIO_USD,RACFACTFIO_1M_USD,RACFACTFIO_1M_3M_USD,RACFACTFIO_3M_1Y_USD,RACFAPPS_USD,RACFAPPS_1M_USD,RACFAPPS_1M_3M_USD,RACFAPPS_3M_1Y_USD,RACFAPPSBP_USD,RACFAPPSBP_1M_USD,RACFAPPSBP_1M_3M_USD,RACFAPPSBP_3M_1Y_USD,RACFAPPSWC_USD,RACFAPPSWC_1M_USD,RACFAPPSWC_1M_3M_USD,RACFAPPSWC_3M_1Y_USD,RACFAPPL_USD,RACFAPPL_1M_USD,RACFAPPL_1M_3M_USD,RACFAPPL_3M_1Y_USD,RACFAPPLBC_USD,RACFAPPLBC_1M_USD,RACFAPPLBC_1M_3M_USD,RACFAPPLBC_3M_1Y_USD,RACFAPPLWP_USD,RACFAPPLWP_1M_USD,RACFAPPLWP_1M_3M_USD,RACFAPPLWP_3M_1Y_USD,RACFAMPAS_USD,RACFAMPAS_1M_USD,RACFAMPAS_1M_3M_USD,RACFAMPAS_3M_1Y_USD,RACFAMPAL_USD,RACFAMPAL_1M_USD,RACFAMPAL_1M_3M_USD,RACFAMPAL_3M_1Y_USD,RACFAMPBS_USD,RACFAMPBS_1M_USD,RACFAMPBS_1M_3M_USD,RACFAMPBS_3M_1Y_USD,RACFAMPBL_USD,RACFAMPBL_1M_USD,RACFAMPBL_1M_3M_USD,RACFAMPBL_3M_1Y_USD,RACFAMPCS_USD,RACFAMPCS_1M_USD,RACFAMPCS_1M_3M_USD,RACFAMPCS_3M_1Y_USD,RACFAMPCL_USD,RACFAMPCL_1M_USD,RACFAMPCL_1M_3M_USD,RACFAMPCL_3M_1Y_USD,RACFAMPDS_USD,RACFAMPDS_1M_USD,RACFAMPDS_1M_3M_USD,RACFAMPDS_3M_1Y_USD,RACFAMPDL_USD,RACFAMPDL_1M_USD,RACFAMPDL_1M_3M_USD,RACFAMPDL_3M_1Y_USD,RACFAMPES_USD,RACFAMPES_1M_USD,RACFAMPES_1M_3M_USD,RACFAMPES_3M_1Y_USD,RACFAMPEL_USD,RACFAMPEL_1M_USD,RACFAMPEL_1M_3M_USD,RACFAMPEL_3M_1Y_USD,RACFAMPFS_USD,RACFAMPFS_1M_USD,RACFAMPFS_1M_3M_USD,RACFAMPFS_3M_1Y_USD,RACFAMPFL_USD,RACFAMPFL_1M_USD,RACFAMPFL_1M_3M_USD,RACFAMPFL_3M_1Y_USD"

IRFCL_MEMORANDUM_ITEMS = "RAMDCD_USD,RAMFIFC_USD,RAMPA_USD,RAMFFS_USD,RAMPAOA_USD,RAMSR_USD,RAMSRLRI_USD,RAMSRLRN_USD,RAMSRBRI_USD,RAMSRBAN_USD,RAMFDA_USD,RAMFDAF_USD,RAMFDAU_USD,RAMFDAW_USD,RAMFDAP_USD,RAMFDAO_USD,RAMFFL_USD,RAMPPS_USD,RAMPPSBP_USD,RAMPPSWC_USD,RAMPPL_USD,RAMPPLBP_USD,RAMPPLWC_USD,RAMCR_USD,RAMCRISDR_USD,RAMCRIC_USD_USD,RAMCRIC_EUR_USD,RAMCRIC_CNY_USD,RAMCRIC_JPY_USD,RAMCRIC_GBP_USD,RAMCROSDR_USD"

IRFCL_TABLES = {
    "reserve_assets_and_other_fx_assets": RESERVE_ASSETS_AND_OTHER_FX_ASSETS,
    "predetermined_drains_on_fx_assets": PREDETERMINED_DRAINS_ON_FX_ASSETS,
    "contingent_drains_fx_assets": CONTINGENT_DRAINS_FX_ASSETS,
    "memorandum_items": IRFCL_MEMORANDUM_ITEMS,
}

IRFCL_PRESET = {
    "irfcl_top_lines": IRFCL_HEADLINE,
    **IRFCL_TABLES,
    "gold_reserves": "RAFAGOLD_USD,RAFAGOLDV_OZT",
    "derivative_assets": "RAMFDA_USD",
}

FREQUENCY_DICT = {
    "month": "M",
    "quarter": "Q",
    "annual": "A",
}
REF_SECTORS_DICT = {
    "government_ex_social_security": "S1311",
    "central_bank": "S121",
    "monetary_authorities": "S1X",
    "all_sectors": "",
}

SECTOR_MAP = {v: k for k, v in REF_SECTORS_DICT.items()}

UNIT_MULTIPLIERS_MAP = {
    "0": "Units",
    "2": "Hundreds",
    "3": "Thousands",
    "6": "Millions",
    "9": "Billions",
    "12": "Trillions",
    "N15": "Quadrillionths",
    "N14": "Hundred Trillionths",
    "N13": "Ten Trillionths",
    "N12": "Trillionths",
    "N11": "Hundred Billionths",
    "N10": "Ten Billionths",
    "N9": "Billionths",
    "N8": "Hundred Millionths",
    "N7": "Ten Millionths",
    "N6": "Millionths",
    "N5": "Hundred Thousandths",
    "N4": "Ten Thousandths",
    "N3": "Thousandths",
    "N2": "Hundredths",
    "N1": "Tenths",
    "1": "Tens",
    "4": "Ten Thousands",
    "5": "Hundred Thousands",
    "7": "Ten Millions",
    "8": "Hundred Millions",
    "10": "Ten Billions",
    "11": "Hundred Billions",
    "13": "Ten Trillions",
    "14": "Hundred Trillions",
    "15": "Quadrillions",
}

TIME_FORMAT_MAP = {
    "P1Y": "Annual",
    "P6M": "Bi-annual",
    "P3M": "Quarterly",
    "P1M": "Monthly",
    "P7D": "Weekly",
    "P1D": "Daily",
}

REF_SECTOR_MAP = {
    "S1311": "Central government excluding social security",
    "S121": "Central Bank",
    "S1X": "Monetary Authorities",
    "1C_AS": "All Sectors",
    "AllSectorsIncludingAllSectors": "All Sectors Including All Sectors",
}


def load_symbols(dataset: str) -> dict:
    """Load IMF symbol list."""
    # pylint: disable=import-outside-toplevel
    import json  # noqa
    from json.decoder import JSONDecodeError
    from pathlib import Path
    from openbb_core.app.model.abstract.error import OpenBBError

    try:
        symbols_file = Path(__file__).parents[1].joinpath("assets", "imf_symbols.json")
        with symbols_file.open(encoding="utf-8") as file:
            symbols = json.load(file)
    except (FileNotFoundError, JSONDecodeError) as e:
        raise OpenBBError(
            f"Failed to load IMF symbols from the static file: {e}"
        ) from e

    if dataset == "all":
        return symbols

    return {k: v for k, v in symbols.items() if v["dataset"] == dataset}



"""IMF Economic Indicators Model."""

# pylint: disable=unused-argument

from datetime import datetime
from typing import Any, Dict, List, Literal, Optional, Union

from openbb_core.app.model.abstract.error import OpenBBError
from openbb_core.provider.abstract.fetcher import Fetcher
from openbb_core.provider.standard_models.economic_indicators import (
    EconomicIndicatorsData,
    EconomicIndicatorsQueryParams,
)
from openbb_core.provider.utils.descriptions import QUERY_DESCRIPTIONS
from openbb_core.provider.utils.errors import EmptyDataError
from openbb_imf.utils.constants import (
    FSI_PRESETS,
    IRFCL_PRESET,
    IRFCL_TABLES,
    load_symbols,
)
from openbb_imf.utils.irfcl_helpers import (
    load_country_to_code_map,
    validate_countries,
)
from pydantic import Field, field_validator


class ImfEconomicIndicatorsQueryParams(EconomicIndicatorsQueryParams):
    """IMF Economic Indicators Query."""

    __json_schema_extra__ = {
        "symbol": {
            "multiple_items_allowed": True,
        },
        "country": {
            "multiple_items_allowed": True,
            "choices": ["all"] + list(list(load_country_to_code_map())),
        },
        "frequency": {
            "choices": ["annual", "quarter", "month"],
        },
    }
    symbol: str = Field(
        default="irfcl_top_lines",
        description=QUERY_DESCRIPTIONS.get("symbol", "")
        + " Use `available_indicators()` to get the list of available symbols."
        + " Use 'IRFCL' to get all the data from International Reserves & Foreign Currency Liquidity indicators."
        + " Use 'core_fsi' to get the core Financial Soundness Indicators."
        + " Use 'core_fsi_underlying' to include underlying data for the core Financial Soundness Indicators."
        + " Complete tables are available only by single country, and are keyed as described below."
        + " The default is 'irfcl_top_lines'. Available presets not listed in `available_indicators()` are:"
        + """\n
        'IRFCL': All the data from the set of indicators. Not compatible with multiple countries.
        'irfcl_top_lines': The default, top line items from the IRFCL data. Compatible with multiple countries.
        'reserve_assets_and_other_fx_assets': Table I of the IRFCL data. Not compatible with multiple countries.
        'predetermined_drains_on_fx_assets': Table II of the IRFCL data. Not compatible with multiple countries.
        'contingent_drains_fx_assets': Table III of the IRFCL data. Not compatible with multiple countries.
        'memorandum_items': The memorandum items table of the IRFCL data. Not compatible with multiple countries.
        'gold_reserves': Gold reserves as value in USD and Fine Troy Ounces. Compatible with multiple countries.
        'derivative_assets': Net derivative assets as value in USD. Compatible with multipile countries.
        'fsi_core': The core Financial Soundness Indicators. Compatible with multiple countries.
        'fsi_core_underlying': The core FSIs underlying series data. Not compatible with country='all'.
        'fsi_encouraged_set': The encouraged set of Financial Soundness Indicators. Not compatible with country='all'.
        'fsi_other': The other Financial Soundness Indicators. Not compatible with country='all'.
        'fsi_balance_sheets': Data categorized as Balance Sheets and Income Statements. Not compatible with country='all'.
        'fsi_all': All the Financial Soundness Indicators. Not compatible with multiple countries.
    """,
    )
    frequency: Literal["annual", "quarter", "month"] = Field(
        default="quarter",
        description="Frequency of the data, default is 'quarter'.",
    )

    @field_validator("symbol", mode="before", check_fields=False)
    @classmethod
    def _count_presets(cls, v):
        """Validate the symbol."""
        if not v:
            return v
        presets = list(IRFCL_PRESET) + FSI_PRESETS
        n_preset = 0
        symbols = v.split(",")
        for symbol in symbols:
            n_preset += 1 if symbol in presets else 0
        if n_preset > 1:
            raise ValueError("only one preset symbol can be used at a time.")
        return v


class ImfEconomicIndicatorsData(EconomicIndicatorsData):
    """IMF Economic Indicators Data."""

    __alias_dict__ = {
        "symbol_root": "parent",
    }

    unit: Optional[str] = Field(
        default=None,
        description="The unit of the value.",
    )
    scale: Optional[str] = Field(
        default=None,
        description="The scale of the value.",
    )
    table: Optional[str] = Field(
        default=None,
        description="The name of the table associated with the symbol.",
    )
    level: Optional[int] = Field(
        default=None,
        description="The indentation level of the data, relative to the table and symbol_root",
    )
    order: Optional[Union[int, float]] = Field(
        default=None,
        description="Order of the data, relative to the table.",
    )
    reference_sector: Optional[str] = Field(
        default=None,
        description="The reference sector for the data.",
    )
    title: Optional[str] = Field(
        default=None,
        description="The title of the series associated with the symbol.",
    )


class ImfEconomicIndicatorsFetcher(
    Fetcher[ImfEconomicIndicatorsQueryParams, List[ImfEconomicIndicatorsData]]
):
    """IMF Economic Indicators Fetcher."""

    @staticmethod
    def transform_query(params: Dict[str, Any]) -> ImfEconomicIndicatorsQueryParams:
        """Transform the query."""
        symbols = params.get("symbol", "")
        countries = params.get("country")
        now = datetime.now().date()
        symbols = (
            "IRFCL"
            if (("all" in symbols or "IRFCL" in symbols) and "fsi_all" not in symbols)
            else symbols if symbols else "irfcl_top_lines"
        )
        incompatible = (
            "fsi_other" in symbols
            or "fsi_encouraged_set" in symbols
            or "fsi_all" in symbols
            or "fsi_core_underlying" in symbols
            or "fsi_balance_sheets" in symbols
        )
        if (symbols == "IRFCL" or incompatible) and not (
            countries or countries == "all"
        ):
            raise OpenBBError(
                f"The selected symbol(s), {params.get('symbol')}, is not compatible with the all-countries group."
                " Please provide country names or two-letter ISO country codes."
            )

        if countries:
            params["country"] = validate_countries(countries)

        if symbols and symbols in IRFCL_PRESET:
            params["symbol"] = IRFCL_PRESET[symbols]
            if symbols in IRFCL_TABLES and countries and countries.split(",") > 1:
                raise OpenBBError(
                    f"Symbol '{symbols}' is a table and can only be used with one country."
                )
        elif symbols:
            params["symbol"] = symbols

        if not params.get("start_date") and (not countries or countries == "all"):
            params["start_date"] = now.replace(
                year=now.year - 1, month=1, day=1
            ).strftime("%Y-%m-%d")

        if not params.get("end_date"):
            params["end_date"] = now.replace(month=12, day=31).strftime("%Y-%m-%d")

        if (not symbols or symbols == "all") and not params.get("start_date"):
            params["start_date"] = now.replace(year=now.year - 1).strftime("%Y-%m-%d")

        return ImfEconomicIndicatorsQueryParams(**params)

    @staticmethod
    async def aextract_data(
        query: ImfEconomicIndicatorsQueryParams,
        credentials: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> List[Dict]:
        """Extract the data."""
        # pylint: disable = import-outside-toplevel
        from openbb_imf.utils.fsi_helpers import _get_fsi_data  # noqa
        from openbb_imf.utils.irfcl_helpers import _get_irfcl_data
        from warnings import warn

        fsi_symbols = load_symbols("FSI")
        irfcl_symbols = load_symbols("IRFCL")
        symbols = query.symbol.split(",")
        new_symbols_irfcl: Union[list, str] = []
        new_symbols_fsi: Union[list, str] = []
        for symbol in symbols:
            if symbol in list(IRFCL_PRESET) + ["all", "IRFCL"]:
                new_symbols_irfcl = symbol
            elif symbol in FSI_PRESETS:
                new_symbols_fsi = symbol
            elif symbol.upper() in fsi_symbols:
                new_symbols_fsi.append(symbol.upper())  # type: ignore
            elif symbol.upper() in irfcl_symbols:
                new_symbols_irfcl.append(symbol.upper())  # type: ignore

        if not new_symbols_irfcl and not new_symbols_fsi:
            raise OpenBBError(
                f"No valid symbols found -> {query.symbol} -> "
                "Use 'available_indicators(provider='imf')' to get the list of available symbols."
            )

        results: list = []
        exceptions: list = []
        try:
            try:
                if new_symbols_irfcl:
                    _kwargs = query.model_dump(exclude_none=True)
                    _kwargs["symbol"] = new_symbols_irfcl
                    results.extend(await _get_irfcl_data(**_kwargs))
            except (EmptyDataError, OpenBBError) as e:
                if new_symbols_fsi:
                    exceptions.append(
                        f"IRFCL dataset error -> {e.__class__.__name__}: {e}"
                    )
                else:
                    raise
            if new_symbols_fsi:
                try:
                    _kwargs = query.model_dump(exclude_none=True)
                    _kwargs["symbol"] = new_symbols_fsi
                    results.extend(await _get_fsi_data(**_kwargs))
                except (EmptyDataError, OpenBBError) as e:
                    if new_symbols_irfcl and len(results) > 0:
                        exceptions.append(
                            f"FSI dataset error -> {e.__class__.__name__}: {e}"
                        )
                    elif not new_symbols_irfcl:
                        raise
        except OpenBBError as exc:
            raise exc from exc

        if not results:
            raise EmptyDataError("No results returned for the query.")

        if results and exceptions:
            msgs = "\n".join(exceptions)
            warn("An error occurred while fetching the data -> " + msgs)

        return results

    @staticmethod
    def transform_data(
        query: ImfEconomicIndicatorsQueryParams,
        data: List[Dict],
        **kwargs: Any,
    ) -> List[ImfEconomicIndicatorsData]:
        """Transform the data."""
        # pylint: disable = import-outside-toplevel
        from numpy import nan
        from pandas import Categorical, DataFrame

        if not data:
            raise EmptyDataError("The data is empty.")

        all_symbols = {
            **load_symbols("all"),
        }
        df = DataFrame(data)

        if df.empty:
            raise EmptyDataError("The data is empty.")

        df = df[df["symbol"].isin(all_symbols)]

        if len(df) == 0:
            raise OpenBBError("The data has a length of 0.")

        df["symbol"] = Categorical(
            df["symbol"],
            categories=all_symbols,
            ordered=True,
        )
        df["parent"] = Categorical(
            df["parent"],
            categories=all_symbols,
            ordered=True,
        )
        df = df.sort_values(
            by=["date", "parent", "symbol", "value"],
            ascending=[True, True, True, False],
        ).reset_index(drop=True)

        df.loc[:, "title"] = df.symbol.apply(
            lambda x: all_symbols.get(x, {}).get("title")
        )
        records = df.replace({nan: None}).to_dict(orient="records")

        return [ImfEconomicIndicatorsData.model_validate(r) for r in records]
